﻿namespace Web.Binding {
    using System;
    using System.CodeDom;
    using System.Collections.Generic;
    using System.Collections.Specialized;
    using System.Linq;
    using System.Text.RegularExpressions;
    using System.Web.UI;

    public class BindableControlBuilder : FileLevelUserControlBuilder {
        private static Regex BindRegex = new Regex(@"{(?<control>\w+)=(?<propertyName>\w+)(,\s*Mode=(?<type>In|Out|TwoWay))?}");

        public override void ProcessGeneratedCode(CodeCompileUnit codeCompileUnit, CodeTypeDeclaration baseType, CodeTypeDeclaration derivedType, CodeMemberMethod buildMethod, CodeMemberMethod dataBindingMethod) {
            // implement IBindableControl
            derivedType.BaseTypes.Add(typeof(IBindableControl));

            var methods = derivedType.Members.OfType<CodeMemberMethod>();

            List<BindingExpression> bindExpressions = new List<BindingExpression>();
            foreach (var method in methods) {
                var bindExpression = GetBindExpression(method);
                if (bindExpression != null) {
                    bindExpressions.Add(bindExpression);
                    if (ShouldGenerateEval(bindExpression)) {
                        GenerateEvalStatement(method, bindExpression);
                    }
                }
            }

            CodeMemberMethod extractValuesMethod = new CodeMemberMethod();
            extractValuesMethod.Name = "ExtractValues";
            extractValuesMethod.ReturnType = new CodeTypeReference(typeof(void));
            extractValuesMethod.Attributes = MemberAttributes.Public;
            extractValuesMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(IOrderedDictionary), "dictionary"));

            foreach (var bindExpression in bindExpressions.Where(ShouldGenerateExtractValues)) {
                GenerateExtractValues(extractValuesMethod, bindExpression);
            }

            derivedType.Members.Add(extractValuesMethod);

            base.ProcessGeneratedCode(codeCompileUnit, baseType, derivedType, buildMethod, dataBindingMethod);
        }

        private bool ShouldGenerateEval(BindingExpression expression) {
            return expression.Mode == BindingMode.In || expression.Mode == BindingMode.TwoWay;
        }

        private bool ShouldGenerateExtractValues(BindingExpression expression) {
            return expression.Mode == BindingMode.Out || expression.Mode == BindingMode.TwoWay;
        }

        private void GenerateExtractValues(CodeMemberMethod extractValuesMethod, BindingExpression bindExpression) {
            // REVIEW: Don't use CodeSnippets so we can support VB
            var extractStatement = new CodeSnippetStatement(String.Format(@"
                var {0} = ({1})FindControl(""{2}"");
                if({0} != null) {{
                    dictionary[""{3}""] = {0}.{4};
                }}
            ", VariableName(bindExpression.ControlID),
             bindExpression.ControlTypeName,
             bindExpression.ControlID,
             bindExpression.PropertyName,
             bindExpression.ControlPropertyName));

            extractValuesMethod.Statements.Add(extractStatement);
        }

        private string VariableName(string name) {
            return Char.ToLower(name[0]) + name.Substring(1);
        }

        private void GenerateEvalStatement(CodeMemberMethod buildControlMethod, BindingExpression bindExpression) {
            // REVIEW: Don't use CodeSnippets so we can support VB
            var evalStatement = new CodeSnippetStatement(String.Format(@"
            @__ctrl.DataBinding += (sender, e) => {{  
                var container = Page.GetDataItem();
                if(container != null) {{
                    (({0})sender).{1} = Convert.ToString(Eval(""{2}""));
                }}   
            }};",
                bindExpression.ControlTypeName,
                bindExpression.ControlPropertyName,
                bindExpression.PropertyName));

            // Put the eval statement before the return
            buildControlMethod.Statements.Insert(buildControlMethod.Statements.Count - 1, evalStatement);
        }

        private BindingExpression GetBindExpression(CodeMemberMethod buildMethod) {
            string controlType = CodeDomUtility.GetControlType(buildMethod);            
            string controlID = CodeDomUtility.GetControlID(buildMethod);            

            var expandoAssignments = from expando in CodeDomUtility.GetExpandoCalls(buildMethod)
                                     let key = (CodePrimitiveExpression)expando.MethodCall.Parameters[0]
                                     let value = (CodePrimitiveExpression)expando.MethodCall.Parameters[1]
                                     select new {
                                         Statement = expando.Statement,
                                         PropertyName = (string)key.Value,
                                         Value = (string)value.Value
                                     };

            foreach (var e in expandoAssignments) {
                if (e.PropertyName == "Binding") {
                    Match match;
                    // Try to match the binding regex and complain if we failed
                    if (!(match = BindRegex.Match(e.Value)).Success) {
                        throw new InvalidOperationException("Invalid bind expression format");
                    }
                    if (String.IsNullOrEmpty(controlID)) {
                        throw new InvalidOperationException("Control ID is required to do databinding!");
                    }
                    // We found a valid binding expression so we are going to remove the statement so we 
                    // don't render it in the page
                    buildMethod.Statements.Remove(e.Statement);                    
                    return new BindingExpression(controlID, controlType, match);
                }
            }
            return null;
        }

        private class BindingExpression {
            public BindingMode Mode { get; private set; }
            public string ControlPropertyName { get; private set; }
            public string PropertyName { get; private set; }
            public string ControlID { get; private set; }
            public string ControlTypeName { get; set; }

            public BindingExpression(string controlID, string controlTypeName, Match m) {
                ControlID = controlID;
                ControlTypeName = controlTypeName;
                Mode = GetMode(m.Groups["type"].Value);
                ControlPropertyName = m.Groups["control"].Value;
                PropertyName = m.Groups["propertyName"].Value;
            }

            private BindingMode GetMode(string type) {
                if (String.IsNullOrEmpty(type)) {
                    return BindingMode.TwoWay;
                }
                return (BindingMode)Enum.Parse(typeof(BindingMode), type);
            }
        }
    }
}